/*
** expr.c - expression evaluation via recursive descent
**
** the grammar is
** EXPR ::= TERM  | TERM BINOP EXPR
** TERM ::= ATOM | EXPR | '(' EXPR ')' | '!' TERM | '-' TERM
** ATOM ::= INTEGER | STRING
**
** constraints :
**   Math operations between strings or between strings and integers are
**   errors
** as in C, relational expressions have the type INTEGER, so you can produce
** silly things like
** 24 + HELLO != GOODBYE
** which returns 25
*/
#include <stdio.h>
#include <ctype.h>
#include <setjmp.h>
/*
** token definitions
*/
#define STRING	-2
#define INTEGER	-3
#define LEXERR	-4
#define NE		-5
#define LE		-6
#define ANDAND	-7
#define GE		-8
#define OROR	-9
#define EQUALSEQUALS -10

/* string and integer terminator tokens */
#define TERMINATORS " \t\n\r\f=!<>&|-+*/%()"
#define HEXVAL(curdigit) \
	(isdigit(curdigit) ? curchar - '0' : (curchar - 'A')+10)
jmp_buf errout;
#define ERROROUT(x) fprintf(stderr,x),longjmp(errout,-1)

/*
** lexval - value associated with STRING or INTEGER tokens
** all expressions resolved via long math, primarily because
** a long integer will always be at least as long as a pointer
*/
static union lval {
	unsigned long ival;
	char *sval;
} lexval;

/*
** static variables used in expression evaluation
*/
static int push_back_token;	/*	if tokens need to be pushed back, 
								they're put here */
static char pushedback = 0;	/* flag indicating a pushed back token */
static char eosflag = 0;	/* true if end of string pushed back */
static char *evalstring;	/* ptr to string being lex'ed */
static int termtype;		/* type of last term */


void main(argc,argv)
char **argv;
{
	long expr(); int base = 10; char goodopt = 0,returnresult = 0;
	long result; char *formatstring = "%ld\n";
	char buf[256]; char *gets();
	if (**++argv == '-')
	{
		while(*(++(*argv)))
		{
			switch(toupper(**argv))
			{
			case 'O':
				base = 010;
				formatstring = "%lo\n";
				goodopt++;
				break;
			case 'H':
				base = 0x10;
				formatstring = "%lx\n";
				goodopt++;
				break;
			case 'D':
				goodopt++;
				formatstring = "%ld\n";
				base = 10;
				break;
			case 'R':
				goodopt++;
				returnresult++;
			}
		}
		if (!goodopt)	/* whoops ... */
		{
			--argv;
		}
		else
			--argc;
	}
	else
		--argv;
	if (argc==1)
	{
		while (NULL != gets(buf))
			printf(formatstring,expr(buf));
	}
	else
	{
		buf[0]='\0';	/* make buffer a null length string */
		while (--argc)
			strcat(buf,*++argv);
		printf(formatstring,result = expr(buf));
	}
	exit(returnresult ? (int)result : 0);
}

/* lexpushback - push a token back into token land.
** You can only push one token back, so it can be assumed that
** the lexval will still be current on the next call to lex
** NOTE - at this point we don't push tokens back
*/
static void lexpushback(token)
int token;
{
	if (pushedback)
	{
		fprintf(stderr,"TOKEN already pushed back\n");
		exit(-1);
	}
	++pushedback;
	push_back_token = token;
}

/*
** lex - get the next token from the input stream
*/
static int lex()
{
	static char buf[128];	/* target of string tokens */
	char *cp = buf;
	extern char *strchr();
	int curchar,nextchar;

	/* if a token has been pushed back return it */
	if (pushedback)
	{	/* the lexval should be unmodified */
		pushedback = 0;
		return push_back_token;
	}
	
	/* skip white space */
	while ('\0' != (curchar = *evalstring++) && isspace(curchar))
		;
	
	if (curchar == '\0')	/* if nothing there return 0 */
		return 0;
	
	/* form of a string is [^0-9=!<>&|-+/%*()][^=!<>&|-+/%*()]* */
	if (NULL == strchr(TERMINATORS "0123456789",curchar))
	{
		/* string token */
		*cp++ = curchar;	/* store first char */
		/* collect rest of string */
		while ((curchar=*evalstring++)!='\0' &&
			NULL == strchr(TERMINATORS,curchar))
				*cp++ = curchar;
		*cp = '\0';			/* terminate the string */
		
		/* push back terminating character */
		--evalstring;
		lexval.sval = buf;
		return STRING;
	}
	/* form of integers is ][0-9]* */
	if (isdigit(curchar))
	{   
		int base = 10;
		if (curchar == '0')
		{
			base = 010;	/* octal */
			curchar = *evalstring++;
			if ('X'==toupper(curchar))
			{
				base = 0x10;	/* hex */
				curchar = *evalstring++;	/* skip 'x' */
			}
		}
		if (base < 0x10)
		{
			lexval.ival = curchar - '0';
			while ((curchar=*evalstring++) != '\0' &&
				isdigit(curchar))
			{
				lexval.ival *= base;
				lexval.ival += curchar - '0';
			}
		} else
		{
			if (!isxdigit(curchar))
				ERROROUT("Bad hex constant!!!\n");
			lexval.ival = HEXVAL(curchar);
			while ((curchar=*evalstring++) != '\0' && isxdigit(curchar))
			{
				curchar = toupper(curchar);
				lexval.ival <<= 4;
				lexval.ival += HEXVAL(curchar);
			}
		}
		/* push back terminating character */
		--evalstring;
		return INTEGER;
	}

	nextchar = *evalstring++;	/* second char of two char tokens */
	switch(curchar)				/* look at first char ... */
	{
	/* look for two character tokens first */
	case '=':
		if (nextchar == '=')
			return EQUALSEQUALS;
		else
			goto return_one_char_token;
	case '!':
		if (nextchar == '=')
			return NE;
		else
			goto return_one_char_token;			
	case '<':
		switch(nextchar)
		{
		case '=':
			return LE;
		case '>':
			return NE;
		default:
			goto return_one_char_token;
		}
	case '>':
		if (nextchar == '=')
			return GE;
		else
			goto return_one_char_token;
	case '&':
		if (nextchar=='&')
			return ANDAND;
		else
			goto return_one_char_token;
	case '|':
		if (nextchar=='|')
			return OROR;
		else
			goto return_one_char_token;
	default:
	/* look for single char tokens */
		if (NULL != strchr("-+*/%()",curchar))
		{
		return_one_char_token:	/* this way out for one char tokens */
			--evalstring;	/* back up a char */
			return curchar;
		}
		/* we bombed - note that the Regular Expression for strings would
		** almost completely eliminate there being invalid tokens - anything
		** unrecognizable would be a string token.
		*/
		return LEXERR;
	}
}

void print_token(lexrval)
{
	switch(lexrval)
	{
	case STRING:
		fprintf(stderr,"STRING : %s",lexval.sval);
		break;
	case INTEGER:
		fprintf(stderr,"INTEGER : %ld",lexval.ival);
		break;
	case NE:
		fprintf(stderr,"NE");
		break;
	case LE:
		fprintf(stderr,"LE");
		break;
	case ANDAND:
		fprintf(stderr,"ANDAND");
		break;
	case GE:
		fprintf(stderr,"GE");
		break;
	case OROR:
		fprintf(stderr,"OROR");
		break;
	case EQUALSEQUALS:
		fprintf(stderr,"EQUALSEQUALS");
		break;
	case LEXERR:
		fprintf(stderr,"LEXERR");
		break;
	case EOF:
		fprintf(stderr,"EOF");
		exit(0);
	case '\n':
		fprintf(stderr,"NEWLINE");
	default:
		fprintf(stderr,"TOKEN : %c ",lexrval);
	}
}

static char *strsave(s)
char *s;
{
	char *rval;
	void *malloc();
	if (NULL == (rval = malloc(strlen(s)+1)))
	{
		fprintf(stderr,"Memory exhausted\n");
		exit(-1);
	}
	strcpy(rval,s);
	return rval;
}

extern long _expr();	/* forward definition */

/*
** term - evaluate term part of grammer - a term being either
** an expression or an operand of a two part expression
*/
static long term()
{
	unsigned long rval; 	/* value returned by term */
#ifdef DEBUG2
	fprintf(stderr,"in term\n");
#endif
	switch(lex())	/* pick up a token */
	{
	case 0:
	case ')':		/* you would see this in an empty string - () - though
					** it would admit silly things like ) being an expression
					*/
		termtype = INTEGER;
		rval = 0L;
		break;
	case '-': /* unary minus */
		rval = term();	/* recurse to pick up term to negate */
		if (termtype == INTEGER)
			rval = -rval;
		else
			ERROROUT("Negation of a string");
		break;
	case '+':	/* unary plus */
		rval = term();
		if (termtype == STRING)
			ERROROUT("Unary plus of a string!!!");
		break;
	case '!':
		rval = term();	/* recurse to pick up logical not */
		if (termtype == INTEGER)
			rval = !rval;
		else
			ERROROUT("Logical negation of a string!!! ");
		break;
	case STRING:
		termtype = STRING;
#ifdef DEBUG2
		fprintf(stderr,"string seen\n");
#endif
		rval = (unsigned long)strsave(lexval.sval);
		break;
	case INTEGER:
		termtype = INTEGER;
#ifdef DEBUG2
		fprintf(stderr,"integer seen\n");
#endif
		rval = lexval.ival;
		break;
	case '(':
		rval = _expr();
		break;
	default:
		fprintf(stderr,stderr,"ERROR : expecting a term, and got ");
		print_token(rval);
		ERROROUT(" ");
	}
#ifdef DEBUG2
	fprintf(stderr,"leaving term - termtype = %s\n",
		termtype == STRING ? "STRING" : "INTEGER");
#endif
	return rval;
}

#define matchterms(type) \
    if (termtype != type) \
		ERROROUT("ERROR : mismatched operands")

/*
** _expr - evaluate expressions part of grammar
*/
static long _expr()
{
	int token;	/* current token */
	unsigned long rval;	/* value returned from _expr */
	int binop;	/* binary operator here */
	unsigned long term1 = 0L,	/* first (or only) term of an expression */
			term2 = 0L;			/* second term of an expression */
#ifdef DEBUG2
	fprintf(stderr,"in expr\n");
#endif
	term1 = term();	/* pick up first term */

	if (0 == (token = lex()) || token == ')') 	/* 0 meaning end of input */
		return term1;	/* simple expression */
	
	/* handle integer expressions */
	if (termtype == INTEGER)
	{
#ifdef DEBUG2
		fprintf(stderr,"Integer seen %d\n",term1);
		fprintf(stderr,"next token = "); print_token(token); putchar('\n');
#endif
		term2 = _expr();
		if (token != ANDAND && token != OROR)
			matchterms(INTEGER);
		switch(token)        
		{
		case '+':                               
			rval = term1 + term2;
			break;
		case '-':
			rval = term1 - term2;
			break;
		case '*':
			rval = term1 * term2;
			break;
		case '/':
			if (0 == term2)
				ERROROUT("ERROR - divide by 0\n");
			rval = term1 / term2;
			break;
		case '%':
			if (0 == term2)
				ERROROUT("ERROR - divide by 0\n");
			rval = term1 % term2;
			break;
		case '|':
			rval = term1 | term2;
			break;
		case '&':
			rval = term1 & term2;
			break;
		case ANDAND:
			rval = term1 && term2;
			break;
		case OROR:
			rval = term1 || term2;
			break;
		case NE:
			rval = term1 != term2;
			break;
		case GE:
			rval = term1 >= term2;
			break;
		case LE:
			rval = term1 <= term2;
			break;
		case '<':
			rval = term1 < term2;
			break;
		case '>':
			rval = term1 > term2;
			break;
		case EQUALSEQUALS:
			rval = term1 == term2;
			break;
		default:
			ERROROUT("Unexpected token\n");
		}
		termtype = INTEGER;
	}
	else
	/* handle string expressions */
	{
#ifdef DEBUG2
		fprintf(stderr,"string seen %s\n",(char *)term1);
		fprintf(stderr,"next token = "); print_token(token); putchar('\n');
#endif
		term2 = _expr();
		matchterms(STRING);
		switch(token)        
		{
		case NE:
			rval = 0 != strcmp((char *)term1,(char *)term2);
			break;
		case GE:
			rval = 0 <= strcmp((char *)term1,(char *)term2);
			break;
		case LE:
			rval = 0 >= strcmp((char *)term1,(char *)term2);
			break;
		case '<':
			rval = 0 > strcmp((char *)term1,(char *)term2);
			break;
		case '>':
			rval = 0 < strcmp((char *)term1,(char *)term2);
			break;
		case EQUALSEQUALS:
			rval = 0 == strcmp((char *)term1,(char *)term2);
			break;
		default:
			ERROROUT("Unexpected token\n");
		}
		termtype = INTEGER;
		free((char *)term1);
		free((char *)term2);
	}
#ifdef DEBUG2
	fprintf(stderr,"leaving _expr, termtype = %s\n",
		termtype == STRING ? "STRING" : "INTEGER");
#endif
	return rval;
}

long expr(buf)
char *buf;
{
	/* initialize statics */
	evalstring = buf;
	pushedback = 0;
	if (-1 == setjmp(errout))
	{
		fprintf(stderr,"Parsing error\n");
		exit(-1);
	}
	return _expr();	/* go do the actual work */
}

/* end of expr.c */
